#include "../btree.h"
#include <gtest/gtest.h>
#include <vector>
#include <string>

#define SMALL_NUM_ENTRIES 100000
#define DEFAULT_NUM_ENTRIES 1000000
#define LARGE_NUM_ENTRIES 10000000

namespace {
  
  static uint32_t db_num_ = 0;

  class BtreeTest : public testing::Test {
  protected:  
    virtual void SetUp()
    {
      bt = new Lux::IO::Btree(Lux::IO::CLUSTER);
      db_name_ = "btc_test";
      num_entries_ = DEFAULT_NUM_ENTRIES;
    }
    virtual void TearDown()
    {
      bt = NULL;
    }
    std::string get_db_name(uint32_t num)
    {
      char buf[256];
      memset(buf, 0, 256);
      sprintf(buf, "%d", num);
      std::string db_name = db_name_ + buf; 
      return db_name;
    }
   
    Lux::IO::Btree *bt; 
    std::string db_name_;
    uint32_t num_entries_;
  };

  /* 
   * operation: put
   * sequence: ordered
   * key: string
   * val: binary
   * omode: create
   **/
  TEST_F(BtreeTest, PutTest) {
    std::string db_name = get_db_name(++db_num_);
    ASSERT_EQ(true, bt->open(db_name.c_str(), Lux::IO::DB_CREAT));

    for (int i = 0; i < num_entries_; ++i) {
      char key[9];
      memset(key, 0, 9);
      sprintf(key, "%08d", i);
      // put
      ASSERT_EQ(true, bt->put(key, strlen(key),
                &i, sizeof(int)));
      // update
      ASSERT_EQ(true, bt->put(key, strlen(key),
                &i, sizeof(int)));
      // no updae
      ASSERT_EQ(true, bt->put(key, strlen(key),
                &i, sizeof(int), Lux::IO::NOOVERWRITE));
    }
    ASSERT_EQ(true, bt->close());
    delete bt;
  }

  /* 
   * operation: get
   * sequence: ordered
   * key: string
   * val: binary
   * omode: rdonly
   **/
  TEST_F(BtreeTest, GetTest) {
    std::string db_name = get_db_name(db_num_);
    ASSERT_EQ(true, bt->open(db_name.c_str(), Lux::IO::DB_RDONLY));

    for (int i = 0; i < num_entries_; ++i) {
      char key[9];
      memset(key, 0, 9);
      sprintf(key, "%08d", i);

      // sequential, system memory
      Lux::IO::data_t *val_data = bt->get(key, strlen(key));
      ASSERT_TRUE(val_data != NULL);
      ASSERT_TRUE(i == *(int *) val_data->data);
      bt->clean_data(val_data);

      memset(key, 0, 9);
      int num = rand() % num_entries_;
      sprintf(key, "%08d", num);

      // random, system memory
      val_data = bt->get(key, strlen(key));
      ASSERT_TRUE(val_data != NULL);
      ASSERT_TRUE(num == *(int *) val_data->data);
      bt->clean_data(val_data);

      memset(key, 0, 9);
      sprintf(key, "%08d", i);
      int val;
      Lux::IO::data_t key_data = {key, strlen(key)};
      Lux::IO::data_t val_data2 = {&val, 0, sizeof(int)};
      Lux::IO::data_t *val_p = &val_data2;

      // sequntial, user memory
      ASSERT_EQ(true, bt->get(&key_data, &val_p));
      ASSERT_TRUE(i == val);
    }
    ASSERT_EQ(true, bt->close());
    delete bt;
  }

  /* 
   * operation: debug print
   * sequence: ordered
   * key: string
   * val: binary
   * omode: create
   **/
  TEST_F(BtreeTest, DebugPrintTest) {
    std::string db_name = get_db_name(db_num_);
    ASSERT_EQ(true, bt->open(db_name.c_str(), Lux::IO::DB_RDONLY));

#ifdef DEBUG
    bt->show_db_header();
    bt->show_root();
#endif

    ASSERT_EQ(true, bt->close());
    delete bt;
  }

  /* 
   * operation: cursor
   * sequence: random
   * key: string
   * val: binary
   * omode: rdonly
   **/
  TEST_F(BtreeTest, CursorTest) {
    std::string db_name = get_db_name(db_num_);
    ASSERT_EQ(true, bt->open(db_name.c_str(), Lux::IO::DB_RDONLY));

    char buf[9] = "00001000";
    Lux::IO::data_t key;
    int num = 1000;
    key.data = buf;
    key.size = 8;
    Lux::IO::cursor_t *c = bt->cursor_init();
    ASSERT_TRUE(c != NULL);
    ASSERT_EQ(true, bt->get(c, &key));
    while (bt->next(c)) {
      Lux::IO::data_t *key, *val;
      ASSERT_EQ(true, bt->cursor_get(c, &key, &val, Lux::IO::SYSTEM));
      ++num;
      ASSERT_EQ(num, *(int *) val->data);

      bt->clean_data(key);
      bt->clean_data(val);
    }
    num = 0;
    ASSERT_EQ(true, bt->first(c));
    do {
      Lux::IO::data_t *key, *val;
      ASSERT_EQ(true, bt->cursor_get(c, &key, &val, Lux::IO::SYSTEM));
      ASSERT_EQ(num, *(int *) val->data);
      ++num;

      bt->clean_data(key);
      bt->clean_data(val);
    } while (bt->next(c));
    num = num_entries_ - 1;
    ASSERT_EQ(true, bt->last(c));
    do {
      Lux::IO::data_t *key, *val;
      ASSERT_EQ(true, bt->cursor_get(c, &key, &val, Lux::IO::SYSTEM));
      ASSERT_EQ(num, *(int *) val->data);
      --num;

      bt->clean_data(key);
      bt->clean_data(val);
    } while (bt->prev(c));
    ASSERT_EQ(true, bt->cursor_fin(c));

    ASSERT_EQ(true, bt->close());
    delete bt;
  }

  /* 
   * operation: del
   * key: string
   * val: binary
   * omode: rdwr
   **/
  TEST_F(BtreeTest, DelTest) {
    std::string db_name = get_db_name(db_num_);
    ASSERT_EQ(true, bt->open(db_name.c_str(), Lux::IO::DB_RDWR));

    for (int i = 0; i < num_entries_; ++i) {
      if (i % 2 == 0) {
        continue;
      }
      char key[9];
      memset(key, 0, 9);
      sprintf(key, "%08d", i);

      Lux::IO::data_t *val_data = bt->get(key, strlen(key));
      ASSERT_TRUE(val_data != NULL);
      bt->clean_data(val_data);
      ASSERT_EQ(true, bt->del(key, strlen(key)));
      val_data = bt->get(key, strlen(key));
      ASSERT_TRUE(bt->get(key, strlen(key)) == NULL);
    }
    ASSERT_EQ(true, bt->close());
    delete bt;
  }

  TEST_F(BtreeTest, PutAgainTest) {
    std::string db_name = get_db_name(db_num_);
    ASSERT_EQ(true, bt->open(db_name.c_str(), Lux::IO::DB_RDWR));
    
    int num = num_entries_ * 5;
    for (int i = num_entries_; i < num; ++i) {
      char key[9];
      memset(key, 0, 9);
      sprintf(key, "%08d", i);
      // put
      ASSERT_EQ(true, bt->put(key, strlen(key),
                &i, sizeof(int)));
    }
    ASSERT_EQ(true, bt->close());
    delete bt;
  }

  TEST_F(BtreeTest, VariableLengthValueTest) {
    std::string db_name = get_db_name(++db_num_);
    ASSERT_EQ(true, bt->open(db_name.c_str(), Lux::IO::DB_CREAT));

    for (int i = 0; i < num_entries_; ++i) {
      char key[9];
      memset(key, 0, 9);
      sprintf(key, "%08d", i);

      char tmp[3];
      memset(tmp, 0, 3);
      sprintf(tmp, "%d", (i % 20)+8);
      std::string format("%0");
      format += tmp;
      format += "d";

      char val[32];
      memset(val, 0, 32);
      sprintf(val, format.c_str(), i); // value is 8 to 27 bytes

      // put
      ASSERT_EQ(true, bt->put(key, strlen(key), val, strlen(val)));

      Lux::IO::data_t *val_data = bt->get(key, strlen(key));
      ASSERT_TRUE(val_data != NULL);
      ASSERT_EQ(strlen(val), val_data->size);
      ASSERT_TRUE(strcmp((char *) val_data->data, val) == 0);
      bt->clean_data(val_data);

      char rval[32];
      memset(rval, 0, 32);
      Lux::IO::data_t key_data = {key, strlen(key)};
      Lux::IO::data_t val_data2 = {rval, 0, 32}; // allocated memory size: 32
      Lux::IO::data_t *val_p = &val_data2;
      ASSERT_EQ(true, bt->get(&key_data, &val_p, Lux::IO::USER));
      ASSERT_TRUE(strcmp((char *) val_p->data, val) == 0);
    }
    ASSERT_EQ(true, bt->close());
    delete bt;
  }

  TEST_F(BtreeTest, VariableLengthValueLargeTest) {
    std::string db_name = get_db_name(++db_num_);
    ASSERT_EQ(true, bt->open(db_name.c_str(), Lux::IO::DB_CREAT));

    for (int i = 0; i < num_entries_; ++i) {
      char key[9];
      memset(key, 0, 9);
      sprintf(key, "%08d", i);

      char tmp[4];
      memset(tmp, 0, 4);
      sprintf(tmp, "%d", (i % 100)+8);
      std::string format("%0");
      format += tmp;
      format += "d";

      char val[128];
      memset(val, 0, 128);
      sprintf(val, format.c_str(), i); // value is 8 to 107 bytes

      // put
      ASSERT_EQ(true, bt->put(key, strlen(key), val, strlen(val)));

      Lux::IO::data_t *val_data = bt->get(key, strlen(key));
      ASSERT_TRUE(val_data != NULL);
      ASSERT_EQ(strlen(val), val_data->size);
      ASSERT_TRUE(strcmp((char *) val_data->data, val) == 0);
      bt->clean_data(val_data);

      char rval[128];
      memset(rval, 0, 128);
      Lux::IO::data_t key_data = {key, strlen(key)};
      Lux::IO::data_t val_data2 = {rval, 0, 128}; // allocated memory size: 128
      Lux::IO::data_t *val_p = &val_data2;
      ASSERT_EQ(true, bt->get(&key_data, &val_p, Lux::IO::USER));
      ASSERT_TRUE(strcmp((char *) val_p->data, val) == 0);
    }
    ASSERT_EQ(true, bt->close());
    delete bt;
  }

  /* 
   * operation: put
   * sequence: random
   * key: string
   * val: binary
   * omode: create
   **/
  TEST_F(BtreeTest, PutTestRandom) {
    std::string db_name = get_db_name(++db_num_);
    ASSERT_EQ(true, bt->open(db_name.c_str(), Lux::IO::DB_CREAT));

    for (int i = 0; i < num_entries_; ++i) {
      char key[9];
      memset(key, 0, 9);
      int num = rand() % num_entries_;
      sprintf(key, "%08d", num);

      ASSERT_EQ(true, bt->put(key, strlen(key), &num, sizeof(int)));
    }
    ASSERT_EQ(true, bt->close());
    delete bt;
  }

  /* 
   * operation: get
   * sequence: random
   * key: string
   * val: binary
   * omode: rdonly
   **/
  TEST_F(BtreeTest, GetTestRandom) {
    std::string db_name = get_db_name(db_num_);
    ASSERT_EQ(true, bt->open(db_name.c_str(), Lux::IO::DB_RDONLY));

    for (int i = 0; i < num_entries_; ++i) {
      char key[9];
      memset(key, 0, 9);
      int num = rand() % num_entries_;
      sprintf(key, "%08d", num);

      Lux::IO::data_t *val_data = bt->get(key, strlen(key));
      if (val_data != NULL) {
        ASSERT_TRUE(num == *(int *) val_data->data);
        bt->clean_data(val_data);
      }
    }
    ASSERT_EQ(true, bt->close());
    delete bt;
  }

  /* 
   * operation: put
   * sequence: ordered
   * key: int
   * val: string
   * omode: create
   **/
  TEST_F(BtreeTest, PutTestKeyInt32) {
    std::string db_name = get_db_name(++db_num_);
    bt->set_cmp_func(Lux::IO::Btree::int32_cmp_func);
    ASSERT_EQ(true, bt->open(db_name.c_str(), Lux::IO::DB_CREAT));

    for (int i = 0; i < num_entries_; ++i) {
      char val[9];
      memset(val, 0, 9);
      sprintf(val, "%08d", i);

      int key = -i;
      ASSERT_EQ(true, bt->put(&key, sizeof(int32_t), val, strlen(val)));
    }
    ASSERT_EQ(true, bt->close());
    delete bt;
  }

  /* 
   * operation: get
   * sequence: random
   * key: binary
   * val: string
   * omode: rdonly
   **/
  TEST_F(BtreeTest, GetTestKeyInt32) {
    std::string db_name = get_db_name(db_num_);
    bt->set_cmp_func(Lux::IO::Btree::int32_cmp_func);
    ASSERT_EQ(true, bt->open(db_name.c_str(), Lux::IO::DB_RDONLY));

    for (int i = 0; i < num_entries_; ++i) {
      char val[9];
      memset(val, 0, 9);
      sprintf(val, "%08d", i);

      int32_t key = -i;
      Lux::IO::data_t *val_data = bt->get(&key, sizeof(int32_t));
      ASSERT_TRUE(val_data != NULL);
      ASSERT_EQ(0, strcmp((char *) val_data->data, val));
      bt->clean_data(val_data);

      Lux::IO::data_t key_data = {&key, sizeof(int32_t)};
      Lux::IO::data_t val_data2 = {val, 0, 9}; // user allocated sizeof(int32_t)
      Lux::IO::data_t *val_p = &val_data2;
      ASSERT_EQ(true, bt->get(&key_data, &val_p, Lux::IO::USER));
      ASSERT_EQ(val_p->size, strlen(val));
      ASSERT_TRUE(strcmp((char *) val_p->data, val) == 0);
    }
    ASSERT_EQ(true, bt->close());
    delete bt;
  }

  TEST_F(BtreeTest, PutTestKeyUint32) {
    std::string db_name = get_db_name(++db_num_);
    bt->set_cmp_func(Lux::IO::Btree::uint32_cmp_func);
    ASSERT_EQ(true, bt->open(db_name.c_str(), Lux::IO::DB_CREAT));

    for (int i = 0; i < num_entries_; ++i) {
      char val[9];
      memset(val, 0, 9);
      sprintf(val, "%08d", i);

      uint32_t key = i;
      ASSERT_EQ(true, bt->put(&key, sizeof(uint32_t), val, strlen(val)));
    }
    ASSERT_EQ(true, bt->close());
    delete bt;
  }

  TEST_F(BtreeTest, GetTestKeyUint32) {
    std::string db_name = get_db_name(db_num_);
    bt->set_cmp_func(Lux::IO::Btree::uint32_cmp_func);
    ASSERT_EQ(true, bt->open(db_name.c_str(), Lux::IO::DB_RDONLY));

    for (int i = 0; i < num_entries_; ++i) {
      char val[9];
      memset(val, 0, 9);
      sprintf(val, "%08d", i);

      uint32_t key = i;
      Lux::IO::data_t *val_data = bt->get(&key, sizeof(uint32_t));
      ASSERT_TRUE(val_data != NULL);
      ASSERT_EQ(0, strcmp((char *) val_data->data, val));
      bt->clean_data(val_data);

      Lux::IO::data_t key_data = {&key, sizeof(uint32_t)};
      Lux::IO::data_t val_data2 = {val, 0, 9}; // user allocated sizeof(int32_t)
      Lux::IO::data_t *val_p = &val_data2;
      ASSERT_EQ(true, bt->get(&key_data, &val_p, Lux::IO::USER));
      ASSERT_EQ(val_p->size, strlen(val));
      ASSERT_TRUE(strcmp((char *) val_p->data, val) == 0);
    }
    ASSERT_EQ(true, bt->close());
    delete bt;
  }

  /* 
   * operation: put, get
   * key: string
   * val: binary
   * omode: create
   **/
  TEST_F(BtreeTest, KeySizeLimitationTest) {
    std::string db_name = get_db_name(++db_num_);
    ASSERT_EQ(true, bt->open(db_name.c_str(), Lux::IO::DB_CREAT));

    char key1[Lux::IO::MAX_KSIZE+1];
    memset(key1, 0, Lux::IO::MAX_KSIZE + 1);
    memset(key1, 'a', Lux::IO::MAX_KSIZE);
    char key2[Lux::IO::MAX_KSIZE+2];
    memset(key2, 0, Lux::IO::MAX_KSIZE + 2);
    memset(key2, 'a', Lux::IO::MAX_KSIZE + 1);
    int val = 10;

    ASSERT_EQ(true, bt->put(key1, strlen(key1), &val, sizeof(int)));
    Lux::IO::data_t *val_data = bt->get(key1, strlen(key1));
    ASSERT_TRUE(val_data != NULL);
    bt->clean_data(val_data);
    ASSERT_EQ(false, bt->put(key2, strlen(key2), &val, sizeof(int)));
    ASSERT_TRUE(bt->get(key2, strlen(key2)) == NULL);
    ASSERT_EQ(false, bt->put(NULL, 0, &val, sizeof(int)));
    ASSERT_TRUE(bt->get((char *) NULL, 0) == NULL);

    ASSERT_EQ(true, bt->close());
    delete bt;
  }

  /* 
   * operation: put, get
   * key: int
   * val: string
   * omode: create
   **/
  TEST_F(BtreeTest, ValueSizeLimitationTest) {
    std::string db_name = get_db_name(++db_num_);
    bt->set_cmp_func(Lux::IO::Btree::int32_cmp_func);
    ASSERT_EQ(true, bt->open(db_name.c_str(), Lux::IO::DB_CREAT));

    int key = 10;
    char val1[Lux::IO::CLUSTER_MAX_VSIZE+1];
    memset(val1, 0, Lux::IO::CLUSTER_MAX_VSIZE + 1);
    memset(val1, 'a', Lux::IO::CLUSTER_MAX_VSIZE);
    char val2[Lux::IO::CLUSTER_MAX_VSIZE+2];
    memset(val2, 0, Lux::IO::CLUSTER_MAX_VSIZE + 2);
    memset(val2, 'a', Lux::IO::CLUSTER_MAX_VSIZE + 1);

    ASSERT_EQ(false, bt->put(&key, sizeof(int), val2, strlen(val2)));
    Lux::IO::data_t *val_data = bt->get(&key, sizeof(int));
    ASSERT_TRUE(val_data == NULL);
    ASSERT_EQ(true, bt->put(&key, sizeof(int), val1, strlen(val1)));
    val_data = bt->get(&key, sizeof(int));
    ASSERT_TRUE(val_data != NULL);
    bt->clean_data(val_data);
    ASSERT_EQ(false, bt->put(&key, sizeof(int), NULL, 0));

    ASSERT_EQ(true, bt->close());
    delete bt;
  }

  /* 
   * sequence: ordered
   * key: string
   * val: binary
   * omode: create
   * concurrency: process
   **/
  TEST_F(BtreeTest, ProcessTest) {
    std::string db_name = get_db_name(++db_num_);
    bt->set_lock_type(Lux::IO::LOCK_PROCESS);
    ASSERT_EQ(true, bt->open(db_name.c_str(), Lux::IO::DB_CREAT));

    for (int i = 0; i < num_entries_; ++i) {
      char key[9];
      memset(key, 0, 9);
      sprintf(key, "%08d", i);
      ASSERT_EQ(true, bt->put(key, strlen(key), &i, sizeof(int)));
    }
    for (int i = 0; i < num_entries_; ++i) {
      char key[9];
      memset(key, 0, 9);
      sprintf(key, "%08d", i);

      Lux::IO::data_t *val_data = bt->get(key, strlen(key));
      ASSERT_TRUE(val_data != NULL);
      ASSERT_TRUE(i == *(int *) val_data->data);
      bt->clean_data(val_data);
    }

    ASSERT_EQ(true, bt->close());
    delete bt;
  }

  /* 
   * sequence: ordered
   * key: string
   * val: binary
   * omode: create
   * concurrency: thread
   **/
  TEST_F(BtreeTest, ThreadTest) {
    std::string db_name = get_db_name(++db_num_);
    bt->set_lock_type(Lux::IO::LOCK_THREAD);
    ASSERT_EQ(true, bt->open(db_name.c_str(), Lux::IO::DB_CREAT));

    for (int i = 0; i < num_entries_; ++i) {
      char key[9];
      memset(key, 0, 9);
      sprintf(key, "%08d", i);
      ASSERT_EQ(true, bt->put(key, strlen(key), &i, sizeof(int)));
    }
    for (int i = 0; i < num_entries_; ++i) {
      char key[9];
      memset(key, 0, 9);
      sprintf(key, "%08d", i);

      Lux::IO::data_t *val_data = bt->get(key, strlen(key));
      ASSERT_TRUE(val_data != NULL);
      ASSERT_TRUE(i == *(int *) val_data->data);
      bt->clean_data(val_data);
    }
    ASSERT_EQ(true, bt->close());
    delete bt;
  }

  /* 
   * operation: put
   * sequence: ordered
   * key: string
   * val: binary
   * omode: create
   * pagesize: 1024
   **/
  TEST_F(BtreeTest, PutTestPageSizeSmallest) {
    std::string db_name = get_db_name(++db_num_);
    bt->set_page_size(1024);
    ASSERT_EQ(true, bt->open(db_name.c_str(), Lux::IO::DB_CREAT));

    for (int i = 0; i < num_entries_; ++i) {
      char key[9];
      memset(key, 0, 9);
      sprintf(key, "%08d", i);
      ASSERT_EQ(true, bt->put(key, strlen(key), &i, sizeof(int)));
    }
    for (int i = 0; i < num_entries_; ++i) {
      char key[9];
      memset(key, 0, 9);
      sprintf(key, "%08d", i);
      Lux::IO::data_t *val_data = bt->get(key, strlen(key));
      ASSERT_TRUE(val_data != NULL);
      ASSERT_TRUE(i == *(int *) val_data->data);
      bt->clean_data(val_data);
    }
    ASSERT_EQ(true, bt->close());
    delete bt;
  }

  /* 
   * operation: put
   * sequence: ordered
   * key: string
   * val: binary
   * omode: create
   * pagesize: 65536
   **/
  TEST_F(BtreeTest, PutTestPageSizeBiggest) {
    std::string db_name = get_db_name(++db_num_);
    bt->set_page_size(65536);
    ASSERT_EQ(true, bt->open(db_name.c_str(), Lux::IO::DB_CREAT));

    for (int i = 0; i < num_entries_; ++i) {
      char key[9];
      memset(key, 0, 9);
      sprintf(key, "%08d", i);
      ASSERT_EQ(true, bt->put(key, strlen(key), &i, sizeof(int)));

    }
    for (int i = 0; i < num_entries_; ++i) {
      char key[9];
      memset(key, 0, 9);
      sprintf(key, "%08d", i);
      Lux::IO::data_t *val_data = bt->get(key, strlen(key));
      ASSERT_TRUE(val_data != NULL);
      ASSERT_TRUE(i == *(int *) val_data->data);
      bt->clean_data(val_data);
    }
    ASSERT_EQ(true, bt->close());
    delete bt;
  }

  /* 
   * operation: put with bulk loading
   * sequence: ordered
   * key: string
   * val: binary
   * omode: create
   **/
  TEST_F(BtreeTest, PutTestWithBulkLoading) {
    std::string db_name = get_db_name(++db_num_);
    ASSERT_EQ(true, bt->open(db_name.c_str(), Lux::IO::DB_CREAT));

    bt->set_bulk_loading(true);
    for (int i = 0; i < num_entries_; ++i) {
      char key[9];
      memset(key, 0, 9);
      sprintf(key, "%08d", i);
      ASSERT_EQ(true, bt->put(key, strlen(key), &i, sizeof(int)));
    }
    bt->set_bulk_loading(false);

    ASSERT_EQ(true, bt->close());
    delete bt;
  }

  /* 
   * operation: cursor lower_bound
   **/
  TEST_F(BtreeTest, CursorLowerBoundTest) {
    std::string db_name = get_db_name(++db_num_);
    ASSERT_EQ(true, bt->open(db_name.c_str(), Lux::IO::DB_CREAT));

    std::vector<std::string> vec;
    vec.push_back("aa");
    vec.push_back("c");
    vec.push_back("e");
    vec.push_back("g");
    vec.push_back("i");
    for (int i = 0; i < vec.size(); ++i) {
      const char *key = vec[i].c_str();
      ASSERT_EQ(true, bt->put(key, strlen(key), &i, sizeof(int)));
    }

    Lux::IO::cursor_t *c = bt->cursor_init();
    ASSERT_TRUE(c != NULL);

    char buf[32] = "a";
    Lux::IO::data_t key;
    key.data = buf;
    key.size = 1;
    ASSERT_EQ(true, bt->lower_bound(c, &key));
    Lux::IO::data_t *k, *v;
    ASSERT_EQ(true, bt->cursor_get(c, &k, &v, Lux::IO::SYSTEM)); 
    ASSERT_TRUE(strcmp((char *) (k->data), "aa") == 0);
    ASSERT_EQ(0, *(int *) (v->data));
    bt->clean_data(k);
    bt->clean_data(v);

    strcpy(buf, "d");
    key.data = buf;
    key.size = 1;
    ASSERT_EQ(true, bt->lower_bound(c, &key));
    ASSERT_EQ(true, bt->cursor_get(c, &k, &v, Lux::IO::SYSTEM)); 
    ASSERT_TRUE(strcmp((char *) (k->data), "e") == 0);
    ASSERT_EQ(2, *(int *) (v->data));
    bt->clean_data(k);
    bt->clean_data(v);

    strcpy(buf, "j");
    key.data = buf;
    key.size = 1;
    ASSERT_EQ(false, bt->lower_bound(c, &key));

    ASSERT_EQ(true, bt->close());
    delete bt;
  }

}

int main(int argc, char *argv[])
{
  testing::InitGoogleTest(&argc, argv); 
  return RUN_ALL_TESTS();
}
